import * as console from "node:console";
import * as crypto from "node:crypto";
import { readFile, writeFile } from "node:fs/promises";
import { join } from "node:path";
import {
	BadRequestException,
	Injectable,
	InternalServerErrorException,
	NotFoundException,
} from "@nestjs/common";
import FileType from "file-type";
import { DbService } from 'apps/backend/src/app/db/db.service';
import { IFileInformation } from 'apps/backend/src/app/storage/storage.types';
import { FilesTypeForMachine, FilesTypesTable } from 'apps/backend/src/app/db/schema';
import { eq } from 'drizzle-orm';

@Injectable()
export class StorageService {
	//TODO make file size configurable by admin
	private maxFileSize = 256; // MiB unit
	constructor(private readonly dbService: DbService) {}


	/**
	 * Save a file to the specified directory.
	 *
	 * @param {string} fileName - The name of the file.
	 * @param {Buffer} file - The file content in Buffer format.
	 * @throws {InternalServerErrorException} If the file save fails.
	 * @return {Promise<void>} A Promise that resolves when the file is successfully saved.
	 */
	private async saveFile(fileName: string, file: Buffer): Promise<void> {
		try {
			await writeFile(join(process.cwd(), "files/", fileName), file, "utf8");
		} catch (err) {
			console.error(err);
			throw new InternalServerErrorException("File save failed !");
		}
	}

	/**
	 * Checks if the current MIME type and file size meet the specified conditions.
	 * @param {Array<string>} machineIds - The IDs of the associated machines.
	 * @param {Buffer} file - The file to check.
	 * @return {Promise<boolean>} - A Promise that resolves to true if the conditions are met, false otherwise.
	 */
	private async checkConditions(machineIds: Array<string>,file: Buffer): Promise<boolean> {

		/**
		 * Checks if the current MIME type is allowed based on the given set of allowed MIME types.
		 * @param {Set<string>} allowedMime - The set of allowed MIME types.
		 * @param {string} currentMime - The current MIME type to check.
		 * @return {boolean} - True if the current MIME type is allowed, false otherwise.
		 */
		function checkMime(allowedMime: Set<string>, currentMime: string): boolean {
			for (const mime of allowedMime) if (mime === currentMime) return true;
			return false;
		}

		const fileType = await FileType.fileTypeFromBuffer(file);

		// Array of MIMEs with possible duplicate field
		const _mimes: Array<string> = []

		// Fetching MIMEs for the associated machines
		for (const machineId of machineIds) {
			console.debug(`Fetching mimeTypes for machine : ${machineId}`)
			// Get MIMEs associated to a machine
			const allowedMimeId = this.dbService.use()
				.select()
				.from(FilesTypeForMachine)
				.where(eq(FilesTypeForMachine.machineId, machineId)).as("allowedMimeId");
			const _allowedMime = await this.dbService.use()
				.select({
					slug: FilesTypesTable.mime,
					name: FilesTypesTable.typeName
				})
				.from(FilesTypesTable)
				.leftJoin(allowedMimeId, eq(FilesTypesTable.id, allowedMimeId.fileTypeId))
			console.debug(`Total : ${_allowedMime.length}`)
			// Append each MIME of a machine
			for (const allowedMimeElement of _allowedMime) {
				_mimes.push(allowedMimeElement.slug)
			}
		}
		//Store the MIMEs without duplicate
		const mimeSet = new Set(_mimes)
		console.debug(`Indexed ${mimeSet.size} unique mimeTypes`)

		//check file size is less than 2mb
		const fileSize = file.byteLength;
		if (fileSize > this.maxFileSize * (1024 * 1024)) {
			throw new BadRequestException(
				"File size to high.",
				{
          cause: "File size",
          description: `File size exceeds the limit. Maximum file size allowed is ${this.maxFileSize}MiB.`
        }
			);
		}

		if (!checkMime(mimeSet, fileType.mime)) throw new BadRequestException(
			{
        cause: "MIME type",
        description: `Invalid MIME type. Allowed MIME types are: ${[...mimeSet].join(", ")}.`
      }
		)

		return true
	}

	/**
	 * Retrieves the contents of a file as a Buffer.
	 * @param {string} fileName - The name of the file to retrieve.
	 * @throws {NotFoundException} If the file is not found.
	 * @returns {Promise<Buffer>} A Promise that resolves to the file contents as a Buffer.
	 */
	private async getFile(fileName: string): Promise<Buffer> {
		try {
			return await readFile(join(process.cwd(), "files/", fileName));
		} catch (err) {
			throw new NotFoundException("File not found");
		}
	}

	/**
	 * Calculates the SHA256 checksum for a given file.
	 *
	 * @param {Buffer} file - The file to calculate the checksum for.
	 * @return {Promise<string>} A Promise resolving to the calculated checksum as a lowercase hexadecimal string.
	 * @throws {InternalServerErrorException} If an error occurs during the checksum calculation.
	 */
	public async getChecksum(file: Buffer): Promise<string> {
		return new Promise((resolve) => {
			try {
				const checksum = crypto.createHash("sha256").update(file).digest("hex").toLowerCase();
				resolve(checksum)
			} catch (err) {
				throw new InternalServerErrorException(err)
			}
		})
	}

	/**
	 * Generate file information based on the provided file and display name.
	 *
	 * @param {Buffer} file - The file in Buffer format.
	 * @param {string} fileDisplayName - The display name of the file.
	 * @param {boolean} [isDocumentation] - Optional flag to indicate if the file is a documentation file.
	 * @returns {Promise<IFileInformation>} - A Promise that resolves to the generated file information.
	 */
	public async generateInformation(file: Buffer, fileDisplayName: string, isDocumentation?: boolean): Promise<IFileInformation> {
		const fileType = await FileType.fileTypeFromBuffer(file);
		const checksum = await this.getChecksum(file)
		const fileName = `${isDocumentation ? "doc" : "file"}-${checksum}.${fileType.ext.toLowerCase()}`;
		return {
			fileName: fileName,
			fileDisplayName: fileDisplayName,
			fileSize: file.byteLength,
			fileChecksum: checksum,
			fileType: fileType,
			isDocumentation: isDocumentation || false
		}
	}

	public async new(fileDisplayName: string, file: Buffer, isDocumentation?: boolean) {
		try {
			const info = await this.generateInformation(file, fileDisplayName, isDocumentation);
			console.log(`Trying to append a new file : "${info.fileDisplayName}"...\n > Checksum SHA-256 : ${info.fileChecksum}\n > Size : ${info.fileSize / (1024*1024)}Mio\n > File format : ${info.fileType.mime}\n`)
			const condition = await this.checkConditions([/* TODO import autorized file format */], file)
			if (!condition) {
				console.warn(`File "${info.fileDisplayName}" did not pass the files requirement.\n${info.fileChecksum}`)
			}

			//TODO Append in DB and save to storage


		} catch (err) {
			throw new BadRequestException(err)
		}
	}
}
